<?php

namespace Ourgold\Exchange\Services;

use Exception;
use Ourgold\Exchange\Models\ExchangeQueue;
use Throwable;

class ExchangeService
{

    /**
     * Поставить вызов метода (методов) класса в очередь
     *
     * @param string[] $methodsChain Одномерный массив имён методов, которые предполагается вызывать цепочкой,
     * то есть каждый следующий метод будет вызываться на результате вызова предыдущего метода
     *
     * @param array|null $arguments Аргументы, которые необходимо передать последнему методу
     *
     * @param string|null $callbackClass FQN класса, реализующий метод __invoke()
     * для обработки результата вызова последнего метода
     *
     * @return ExchangeQueue
     *
     * @see \Ourgold\Exchange\Services\ExternalAPIs
     * @see self::processQueueEntry
     * @see self::processQueue
     */
    public static function queue(
        array   $methodsChain,
        ?array  $arguments = null,
        ?string $callbackClass = null
    ): ExchangeQueue
    {

        self::validateConfig();

        $data = [
            'methodsChain' => $methodsChain,
            'arguments' => $arguments,
            'callbackClass' => $callbackClass,
        ];

        $storeService = new ExchangeQueueStoreService();
        return $storeService->createRaw($data);
    }

    /**
     * Выполнить задания из очереди обмена
     */
    public function processQueue()
    {
        self::validateConfig();

        $delay = intval(config('exchange.delay'));

        $storeService = new ExchangeQueueStoreService();
        $items = $storeService->getUnprocessed();
        foreach ($items as $item) {
            $error = null;

            try {
                $this->processQueueEntry($item);
            } catch (Throwable $e) {
                $error = $e->getMessage();
            }

            $storeService->update($item->id, $error);
            sleep($delay);
        }
    }

    /**
     * @throws Exception
     */
    public function processQueueEntry(ExchangeQueue $queueEntry)
    {
        if (
            !isset($queueEntry->data['methodsChain'])
            || empty($queueEntry->data['methodsChain'])
        ) {
            throw new Exception("'methodsChain' not found or empty");
        }

        $superclassName = config('exchange.api_superclass');

        /**
         * @var string[] $methodsChain
         * @var array $arguments
         */
        $apiSuperClass = new $superclassName;
        $methodsChain = $queueEntry->data['methodsChain'];
        $arguments = $queueEntry->data['arguments'] ?? [];

        $numberOfMethods = count($methodsChain);

        $lastResult = null;

        foreach ($methodsChain as $i => $method) {

            // Если результат возврата предыдущего метода - объект (инстанс класса),
            // то вызываем следующий метод относительно него, получая цепочку
            $classToCallMethodAt = $lastResult ?? $apiSuperClass;

            // Передаем аргументы только последнему методу в цепочке
            if ($i === $numberOfMethods - 1) {
                $lastResult = call_user_func([$classToCallMethodAt, $method], ...$arguments);
            } else {
                $lastResult = call_user_func([$classToCallMethodAt, $method]);
            }
        }

        // Если есть результат вызова и указан класс-коллбэк, пытаемся вызвать его
        if (isset($lastResult, $queueEntry->data['callbackClass'])) {
            $callbackClassName = $queueEntry->data['callbackClass'];

            if (!class_exists($callbackClassName)) {
                \Log::error("class $callbackClassName does not exist!");
            } else {
                $callbackClass = new $callbackClassName;
                return $callbackClass($lastResult);
            }
        }

        return $lastResult;
    }

    /**
     * @throws Exception
     */
    private static function validateConfig()
    {
        if (config('exchange.api_superclass') === ExternalAPIs::class) {
            throw new Exception("Default superclass! Publish config and change it to your own superclass FQN");
        }
    }
}
